import nrrd
import pickle
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt 

import os
basepath = os.path.abspath(__file__)
folder = os.path.dirname(basepath)


ANO, metaANO = nrrd.read(folder+'/ABAanno/ccfv3/annotation_25.nrrd')
onto1_r = pickle.load(open(folder+'/ABAanno/ontology1_reverse.pkl', 'rb'))
onto2_r = pickle.load(open(folder+'/ABAanno/ontology2_reverse.pkl', 'rb'))
id_name = pd.read_csv(folder+'/ABAanno/ontology.csv', index_col=0)['name'].to_dict()
name_color = pd.read_csv(folder+'/ABAanno/ontology.csv', index_col=0).set_index('name')['color_hex_triplet'].to_dict()


# get index of annotation array from CCF coordinates.
def get_index(coord):
    ML_ref, DV_ref, AP_ref = coord[:,0], coord[:,1], coord[:,2]
    ML = ML_ref*40 + 228
    DV = -DV_ref*40
    AP = 214 - AP_ref*40
    AP [AP > 527] = 527
    DV [DV > 319] = 319
    ML [ML > 455] = 455
    return np.vstack((AP, DV, ML)).T.astype(int).tolist()



def ABA_anno(adata, ML_key, DV_key, AP_key, spatial_key='spatial'):
 
    """
    
    Get ABA annotation based on CCF coordinates
    
    Parameters
    ----------
    adata
        AnnData object of scanpy package
    ML_key
        The key of medial/lateral (ML) coordinates in .obs
    DV_key
        The key of dorsal/ventral (DV) coordinates in .obs
    AP_key
        The key of anterior/posterior (AP) coordinates in .obs
    spatial_key
        The key of 2D coordinates in .obsm

    """
    
    coord = adata.obs[[ML_key, DV_key, AP_key]].values  # get 3D coordinates
    index_ABA = get_index(coord)   # get 3D index in annotation array

    annos = []  # annotation of each spots
    for index_tmp in index_ABA:
        index_ap, index_dv, index_ml = index_tmp
        annos.append(ANO[index_ap, index_dv, index_ml])

    annos = np.array(annos)

    # refine annotation spots with invalid annotaion 0 or 997.
    inval_index = np.array(np.where(annos==0)[0].tolist() + np.where(annos==997)[0].tolist())

    if len(inval_index) != 0:
        coord_2d = adata.obsm[spatial_key]
        cord_query = coord_2d[inval_index]

        from sklearn.neighbors import NearestNeighbors
        neigh = NearestNeighbors(n_neighbors=50)
        neigh.fit(coord_2d)
        indices = neigh.kneighbors(cord_query, return_distance=False)
        indices = indices[:,1:]

        annos_inval_ = annos[indices]
        annos_inval = [np.setdiff1d(annos_inval_[i], [0, 997])[0] for i in range(len(annos_inval_))]
        annos[inval_index] = annos_inval

    # define annotaions and the cooresponding colors in ABA in 2 levels.
    adata.obs['ABA_id'] = annos
    adata.obs['ABA_name_level2'] = adata.obs['ABA_id'].replace(onto2_r).replace(id_name).astype('category')
    adata.obs['ABA_name_level1'] = adata.obs['ABA_id'].replace(onto2_r).replace(onto1_r).replace(id_name).astype('category')

    adata.obs['ABA_color_level2'] = adata.obs['ABA_name_level2'].replace(name_color)
    adata.obs['ABA_color_level1'] = adata.obs['ABA_name_level1'].replace(name_color)
    
    return adata


def plot_spatial_ABA(adatas, spatial_key, level='level2', title_key='section_index', 
                     s=10, ncols=4, figsize=(5.5, 3), legend=False, save='spatial.png'):
    """
    
    Spatial plot with ABA color maps.
    
    Parameters
    ----------
    adatas
        List of AnnData object of scanpy package
    spatial_key
        The key of 2D coordinates in .obsm
    level
        The annotation lavel of ABA
    """
    
    import seaborn as sns
    import matplotlib.patches as mpatches
    if isinstance(s, list):
        assert len(adatas) == len(s)
    else:
        s = [s]*len(adatas)  
    level_use = 'ABA_name_' + level
    color_use = 'ABA_color_' + level 
    if len(adatas) < ncols:
        ncols = len(adatas)
        nrows = 1
    else:
        nrows = int(np.ceil(len(adatas)/ncols))
    if (nrows == 1) & (ncols == 1):
        plt.figure(figsize=figsize)
        sns.scatterplot(data = adatas[0].obs,
                        x = 'stereo_ML',
                        y = 'stereo_DV',
                        label = level_use, 
                        color = adatas[0].obs[color_use], 
                        s = s[0],
                        legend=legend)
        if legend:
            plt.legend( handles=[mpatches.Patch(color=name_color[name_tmp], label=name_tmp) for name_tmp in adata.obs[level_use].cat.categories.tolist()], 
                            bbox_to_anchor=(1.0,1.0),
                            fontsize=5,
                            title_fontsize=5)
        plt.axis('off')
        plt.title(adatas[0].obs[title_key][0])
        plt.savefig(save, bbox_inches='tight', dpi=300)
        plt.show()
        plt.close()
    else:
        fig, axs = plt.subplots(nrows, ncols, figsize=figsize,constrained_layout=True)   
        if nrows == 1:  
            for i in range(len(adatas)):
                axs[i].get_xaxis().set_visible(False)
                axs[i].get_yaxis().set_visible(False)
                axs[i].axis('off')
                axs[i].set_xlim([adatas[i].obsm[spatial_key].min(0)[0], adatas[i].obsm[spatial_key].max(0)[0]])
                axs[i].set_ylim([adatas[i].obsm[spatial_key].min(0)[1], adatas[i].obsm[spatial_key].max(0)[1]])
                sns.scatterplot(data = adatas[i].obs,
                                x = 'stereo_ML',
                                y = 'stereo_DV',
                                label = level_use, 
                                color = adatas[i].obs[color_use], 
                                s = s[i], 
                                ax=axs[i],
                                legend=legend)
                axs[i].set_title(adatas[i].obs[title_key][0])
                if legend:
                    if i == len(adatas)-1:
                        plt.legend( handles=[mpatches.Patch(color=name_color[name_tmp], label=name_tmp) for name_tmp in adata.obs[level_use].cat.categories.tolist()], 
                                    bbox_to_anchor=(1.0,1.0),
                                    fontsize=5,
                                    title_fontsize=5)
            fig.savefig(save, bbox_inches='tight', dpi=300)
            plt.show()
            plt.close()
        else:
            index = 0
            for i in range(nrows):
                for j in range(ncols):
                    axs[i,j].get_xaxis().set_visible(False)
                    axs[i,j].get_yaxis().set_visible(False)
                    axs[i,j].axis('off')
                    axs[i,j].set_xlim([adatas[index].obsm[spatial_key].min(0)[0], adatas[index].obsm[spatial_key].max(0)[0]])
                    axs[i,j].set_ylim([adatas[index].obsm[spatial_key].min(0)[1], adatas[index].obsm[spatial_key].max(0)[1]])
                    axs[i,j].set_title(adatas[index].obs[title_key][0])
                    if index < len(adatas):
                        key = keys_use[index]
                        if j<(n_col-1):
                            sns.scatterplot(data = adatas[index].obs,
                                            x = 'stereo_ML',
                                            y = 'stereo_DV',
                                            label = level_use, 
                                            color = adatas[index].obs[color_use], 
                                            s = s[index], 
                                            ax=axs[i,j],
                                            legend=legend)
                            j += 1 
                        else:
                            sns.scatterplot(data = adatas[index].obs,
                                            x = 'stereo_ML',
                                            y = 'stereo_DV',
                                            label = level_use, 
                                            color = adatas[index].obs[color_use], 
                                            s = s[index], 
                                            ax=axs[i,j],
                                            legend=legend)
                            if legend:
                                plt.legend( handles=[mpatches.Patch(color=name_color[name_tmp], label=name_tmp) for name_tmp in adata.obs[level_use].cat.categories.tolist()], 
                                            bbox_to_anchor=(1.0,1.0),
                                            fontsize=5,
                                            title_fontsize=5)
                            i += 1
                            j = 0       
                    index += 1
            fig.savefig(save, bbox_inches='tight', dpi=300)
            plt.show()
            plt.close()








import numpy as np 
from sklearn.neighbors import NearestNeighbors

from .transformation import best_fit_transform, nearest_neighbor, transform


def align_init_pair(adata1, 
                    adata2, 
                    spatial_key1,
                    spatial_key2 = None,
                    num_mnn = 1,
                    emb_key = 'HAN_SE',
                    key_added = 'transform_init'
                    ):
    '''
    Align spatial coordinates of adata2 to match adata1 according to the similarity of emb_key.
    Input:
      adata1
      adata2
      spatial_key1: Key of spatial coordinates in .obsm 
      spatial_key2: Key of spatial coordinates in .obsm
      num_mnn:      The number of mutual nearest neighbors calculated according to emb_key
      emb_key:      The key used to calculate spot similarity in .obsm
      key_added:    Key of transformed coordinates added in .obsm
    Returns:
      adata1 with 'transformed' in .obsm
      adata2 with 'transformed' in .obsm
      pair: np.array of indices of nearest neighbor pairs
      T: transformation matrix with 3 x 3
    '''
    
    print(f'Finding similar pairs using {emb_key}...')
    
    # Extract features used to calculate spot similarity
    emb1_tmp = adata1.obsm[emb_key]
    emb2_tmp = adata2.obsm[emb_key]
    
    # Search for mutual nearest neighbors of two slices
    neigh1 = NearestNeighbors(n_neighbors=num_mnn, metric='cosine')
    neigh1.fit(emb2_tmp)
    indices1 = neigh1.kneighbors(emb1_tmp, return_distance=False)
    neigh2 = NearestNeighbors(n_neighbors=num_mnn, metric='cosine')
    neigh2.fit(emb1_tmp)
    indices2 = neigh2.kneighbors(emb2_tmp, return_distance=False)
    
    set1 = {(i, indices1[i,j]) for i in range(indices1.shape[0]) for j in range(indices1.shape[1])}
    set2 = {(indices2[j,i], j) for j in range(indices2.shape[0]) for i in range(indices2.shape[1])}
    
    pair = set1.intersection(set2)
    pair = np.array(list(pair))
    
    print(f'Aligning slices using {pair.shape[0]} pairs of similar spots!')
    
    # Extract the coordinates of the nearest neighbors
    if spatial_key2 is None:
        spatial_key2 = spatial_key1
    
    B = adata1[pair[:,0],:].obsm[spatial_key1]
    A = adata2[pair[:,1],:].obsm[spatial_key2]
    
    # search for best transformation
    T,_,_ = best_fit_transform(A, B)
    
    # transform the coordinates of adata2
    adata2.obsm[key_added] = transform(adata2.obsm[spatial_key2], T)    
    
    if key_added!=spatial_key1:
        adata1.obsm[key_added] = adata1.obsm[spatial_key1]
    
    return adata1, adata2, pair, T


# def align_init_pair(adata1, 
#                     adata2, 
#                     spatial_key1,
#                     spatial_key2 = None,
#                     num_mnn = 1,
#                     emb_key = 'HAN_SE',
#                     key_added = 'transform_init',
#                     init_pose = None, 
#                     max_iterations = 50, 
#                     tolerance = 0.001
#                     ):
#     '''
#     Align spatial coordinates of adata2 to match adata1 according to the similarity of emb_key.
#     Input:
#       adata1
#       adata2
#       spatial_key1: Key of spatial coordinates in .obsm 
#       spatial_key2: Key of spatial coordinates in .obsm
#       num_mnn:      The number of mutual nearest neighbors calculated according to emb_key
#       emb_key:      The key used to calculate spot similarity in .obsm
#       key_added:    Key of transformed coordinates added in .obsm
#     Returns:
#       adata1 with 'transformed' in .obsm
#       adata2 with 'transformed' in .obsm
#       pair: np.array of indices of nearest neighbor pairs
#       T: transformation matrix with 3 x 3
#     '''
    
#     print(f'Finding similar pairs using {emb_key}...')
    
#     # Extract features used to calculate spot similarity
#     emb1_tmp = adata1.obsm[emb_key]
#     emb2_tmp = adata2.obsm[emb_key]
    
#     # Search for mutual nearest neighbors of two slices
#     neigh1 = NearestNeighbors(n_neighbors=num_mnn, metric='cosine')
#     neigh1.fit(emb2_tmp)
#     indices1 = neigh1.kneighbors(emb1_tmp, return_distance=False)
#     neigh2 = NearestNeighbors(n_neighbors=num_mnn, metric='cosine')
#     neigh2.fit(emb1_tmp)
#     indices2 = neigh2.kneighbors(emb2_tmp, return_distance=False)
    
#     set1 = {(i, indices1[i,j]) for i in range(indices1.shape[0]) for j in range(indices1.shape[1])}
#     set2 = {(indices2[j,i], j) for j in range(indices2.shape[0]) for i in range(indices2.shape[1])}
    
#     pair = set1.intersection(set2)
#     pair = np.array(list(pair))
    
#     print(f'Aligning slices using {pair.shape[0]} pairs of similar spots!')
    
#     # Extract the coordinates of the nearest neighbors
#     if spatial_key2 is None:
#         spatial_key2 = spatial_key1
    
#     B = adata1[pair[:,0],:].obsm[spatial_key1]
#     A = adata2[pair[:,1],:].obsm[spatial_key2]
    
#    # get number of dimensions
#     m = A.shape[1]

#     # make points homogeneous, copy them to maintain the originals
#     src = np.ones((m+1,A.shape[0]))   # (3, 238)
#     dst = np.ones((m+1,B.shape[0]))   # (3, 248)
#     src[:m,:] = np.copy(A.T)
#     dst[:m,:] = np.copy(B.T)

#     # apply the initial pose estimation
#     if init_pose is not None:
#         src = np.dot(init_pose, src)

#     prev_error = 0
#     for i in range(max_iterations): 
        
#         # find the nearest neighbors between the current source and destination points
#         distances, indices = nearest_neighbor(src[:m,:].T, dst[:m,:].T) 
        
#         # compute the transformation between the current source and nearest destination points
#         T,_,_ = best_fit_transform(src[:m,:].T, dst[:m,indices].T)
        
#         # update the current source
#         src = np.dot(T, src)
        
#         # check error
#         mean_error = np.mean(distances)
#         print('iter', i, 'error:', np.abs(prev_error - mean_error))
        
#         if np.abs(prev_error - mean_error) < tolerance:
#             break
        
#         prev_error = mean_error
        
#     # calculate final transformation
#     T,_,_ = best_fit_transform(A, src[:m,:].T)

#     if key_added != spatial_key1:
#         adata1.obsm[key_added] = adata1.obsm[spatial_key1]
    
#     adata2.obsm[key_added] = transform(adata2.obsm[spatial_key2], T)
    
#     return adata1, adata2, pair, T


def initial_alignment(  adatas, 
                        spatial_key = 'spatial',
                        emb_key = 'HAN_SE',
                        num_mnn = 1, 
                        key_added = 'transform_init'
                        ):
    
    '''
    Perform initial alignment for adatas according to the similarity of emb_key.
    Input:
      adatas:       Adata list to be aligned
      spatial_key:  Key of spatial coordinates in .obsm 
      emb_key:      The key used to calculate spot similarity in .obsm
      num_mnn:      The number of mutual nearest neighbors calculated according to emb_key
      key_added:    Key of transformed coordinates added in .obsm
    Returns:
      adatas:       Aligned adata list with 'transformed' in .obsm
      anchors:      np.array of indices of nearest neighbor pairs
    '''
    
    aligned_init = []
    anchors = []
    Ts = []
    
    print('Performing initial alignment of the 1 pair of data...')
    
    # initial alignment of the first two adatas
    adata1, adata2 = adatas[0], adatas[1]
    
    adata1, adata2, anchor_tmp, T_tmp = align_init_pair(adata1, 
                                                        adata2, 
                                                        spatial_key1 = spatial_key,
                                                        spatial_key2 = None,
                                                        num_mnn = num_mnn, 
                                                        emb_key = emb_key,
                                                        key_added = key_added
                                                        )
    
    aligned_init.append(adata1)
    aligned_init.append(adata2)
    anchors.append(anchor_tmp)
    Ts.append(T_tmp)
    
    # initial alignment of the rest adatas
    if len(aligned_init)!=len(adatas):
        for i in range(1, len(adatas)-1):
            
            print(f'Performing initial alignment of the {i+1} pair of data...')
            
            adata1_tmp = aligned_init[-1]
            adata2_tmp = adatas[i+1]
            
            for T_ in Ts:
                adata2_tmp.obsm[spatial_key] = transform(adata2_tmp.obsm[spatial_key], T_)
            
            _, adata2_tmp, anchor_tmp, T_tmp = align_init_pair( adata1_tmp, 
                                                                adata2_tmp, 
                                                                spatial_key1 = key_added, 
                                                                spatial_key2 = spatial_key, 
                                                                num_mnn = num_mnn, 
                                                                emb_key = emb_key,
                                                                key_added = key_added
                                                               )
                    
            aligned_init.append(adata2_tmp)
            anchors.append(anchor_tmp)
            Ts.append(T_tmp)
    
    return aligned_init, anchors, Ts




def initial_alignment_( adatas, 
                        spatial_key = 'spatial',
                        emb_key = 'HAN_SE',
                        num_mnn = 1, 
                        key_added = 'transform_init'
                        ):
    '''
    Perform initial alignment for adatas according to the similarity of emb_key.
    Input:
      adatas:       Adata list to be aligned
      spatial_key:  Key of spatial coordinates in .obsm 
      emb_key:      The key used to calculate spot similarity in .obsm
      num_mnn:      The number of mutual nearest neighbors calculated according to emb_key
      key_added:    Key of transformed coordinates added in .obsm
    Returns:
      adatas:       Aligned adata list with 'transformed' in .obsm
      anchors:      np.array of indices of nearest neighbor pairs
    '''
    
    Ts_init = []
    for i in range(len(adatas)-1):
        adata1 = adatas[i]
        adata2 = adatas[i+1]
        _, _, _, T_ = align_init_pair(  adata1, 
                                        adata2, 
                                        spatial_key1 = spatial_key,
                                        spatial_key2 = None,
                                        num_mnn = num_mnn,
                                        emb_key = emb_key,
                                        key_added = key_added
                                        )
        Ts_init.append(T_)
    # Compute initial transformed coordinates
    for i in range(len(adatas)):          
        adatas[i].obsm[key_added] = adatas[i].obsm[spatial_key]
        if i != 0:
            T_tmp = Ts_init[:i]
            for T_ in T_tmp:
                adatas[i].obsm[key_added] = transform(adatas[i].obsm[key_added], T_)
    return adatas, Ts_init







import pandas as pd
import numpy as np


def calcu_lisi(adata, domain_key, spatial_key):
    '''
    Calculate index LISI of given cluster and spatial coordinates;
    Input:  adata, domain_key in .obs, spatial_keyin .obsm;
    Output: list of sorted domains
    '''
    import rpy2.robjects as ro
    ro.r.library("lisi")
    ro.r['set.seed'](1)
    
    from rpy2.robjects import pandas2ri
    pandas2ri.activate()
    
    lisi = ro.r['compute_lisi']
    result = adata.obs[[domain_key]]
    result[['x', 'y']] = adata.obsm[spatial_key].copy()
    
    res = lisi(ro.conversion.py2rpy(result[['x', 'y']]), 
                ro.conversion.py2rpy(result[[domain_key]]), 
                ro.StrVector([domain_key]))
    
    result['lisi'] = pd.DataFrame(res).astype(float).values.flatten()
    
    return result


def select_clustered_domains(lisi1, lisi2, domain_key, use_domain_nums=3, sep_sort=True):
    '''
    Sort domains according to the index LISI of each domain;
    Input:  lisi1: DataFrame of data1 with columns: ['lisi', domain_key]
            lisi2: DataFrame of data2 with columns: ['lisi', domain_key]
            domain_key;
            use_domain_num: number of domains used to aligning slices.
            sep_sort:  Boolean value, whether to sort spatial clustered pattern together.
    Output: list of used domains
    '''
    
    if sep_sort:
        
        # select domains with more than 3 points
        domain_ok1 = lisi1[domain_key].value_counts()[lisi1[domain_key].value_counts() > 3].index.tolist()
        domain_ok2 = lisi2[domain_key].value_counts()[lisi2[domain_key].value_counts() > 3].index.tolist()

        # sort spatial clustering pattern of domains according to index LISI
        domains_sorted1 = lisi1.groupby(domain_key).median().loc[domain_ok1,:].sort_values('lisi').index.tolist()
        domains_sorted2 = lisi2.groupby(domain_key).median().loc[domain_ok2,:].sort_values('lisi').index.tolist()
        
        
        anchor1_dict = {domains_sorted1[ii]:ii for ii in range(len(domains_sorted1))}
        anchor2_dict = {domains_sorted2[ii]:ii for ii in range(len(domains_sorted2))}
        
        # common domains exist in two slices
        common_domains = list(set(anchor1_dict.keys()).intersection(set(anchor2_dict.keys())))
        
        # check the number of domains and domains needed
        if len(common_domains) < use_domain_nums:
            print('The number of regions required is greater than the actual number of regions')
        
        # sort domains based on two slices 
        anchor_dict = {ii: anchor1_dict[ii] + anchor2_dict[ii] for ii in common_domains}
        domains_sorted = sorted(anchor_dict.items(), key = lambda kv:(kv[1], kv[0]))
        domains_sorted = [ii[0] for ii in domains_sorted]
        domain_use = domains_sorted[:use_domain_nums]
    
    else:
        
        # common domains exist in two slices
        common_domains = list(set(lisi1[domain_key]).intersection(set(lisi2[domain_key])))
        
        # check the number of domains and domains needed
        if len(common_domains) < use_domain_nums:
            print('The number of regions required is greater than the actual number of regions')
        
        lisi_tmp = pd.concat([lisi1, lisi2])
        lisi_tmp = lisi_tmp[lisi_tmp[domain_key].isin(common_domains)]
        
        # sort spatial clustering pattern of domains according to index LISI
        domains_sorted = lisi_tmp.groupby(domain_key).median().sort_values('lisi').index.tolist()
        domain_use = domains_sorted[:use_domain_nums]
    
    return domain_use



from typing import List
from itertools import chain
from scipy.spatial import Delaunay

def alpha_shape(points, alpha, only_outer=True)->List:
    '''
    Compute the alpha shape (concave hull) of a set of points.  
    Parameters
    ----------
    points
        np.array of shape (n,2) points.
    alpha
        alpha value.
    only_outer
    boolean value to specify if we keep only the outer border or also inner edges.
    Return
    ----------
    Set of (i,j) pairs representing edges of the alpha-shape. (i,j) are the indices in the points array.
    Refer
    ----------
    https://stackoverflow.com/questions/50549128/boundary-enclosing-a-given-set-of-points
    '''
    assert points.shape[0] > 3, "Need at least four points"
    def add_edge(edges, i, j):
        """
        Add an edge between the i-th and j-th points,
        if not in the list already
        """
        if (i, j) in edges or (j, i) in edges:
            # already added
            assert (j, i) in edges, "Can't go twice over same directed edge right?"
            if only_outer:
                # if both neighboring triangles are in shape, it's not a boundary edge
                edges.remove((j, i))
            return
        edges.add((i, j))
    tri = Delaunay(points)
    edges = set()
    # Loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    circum_r_list = []
    for ia, ib, ic in tri.simplices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]
        # Computing radius of triangle circumcircle
        # www.mathalino.com/reviewer/derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle
        a = np.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
        b = np.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
        c = np.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)
        s = (a + b + c) / 2.0
        area = np.sqrt(s * (s - a) * (s - b) * (s - c))
        circum_r = a * b * c / (4.0 * area)
        circum_r_list.append(circum_r)
        if circum_r < alpha:
            add_edge(edges, ia, ib)
            add_edge(edges, ib, ic)
            add_edge(edges, ic, ia)
    boundary = list(set(list(chain.from_iterable(list(edges)))))
    return boundary, edges, circum_r_list



from operator import itemgetter

def detect_edge_of_domains(adatas, domain_key, domains_use, spatial_key='spatial', alpha=50):
    '''
    Detect edges which were used to aligning the slices, including edges of slices and used domains.
    Parameters
    ----------
    adatas
        List of adata.
    domain_key
        Key of spatial domains in .obs 
    domains_use
        Domains used to align slices calculated 'by select_clustered_domains'. list with length len(adatas)-1.
    spatial_key
        Key of spatial coordinates in .obsm
    alpha
        alpha value.
    plot_edges
        Boolean value of whether to plot the used edges or not.
    result_path
        Path to save the used edges if plot_edges=True.
    Return
    ----------
    boundary_all
        List of boundary indexs in each slices 
    edge_all
        List of edge indexs in each slices 
    '''

    assert len(domains_use) == len(adatas)-1
    
    boundary_all = []
    edge_all = []
    
    for i in range(len(domains_use)): 

        domain_use = domains_use[i]
        adata1_tmp = adatas[i]
        adata2_tmp = adatas[i+1]    
        
        # Detect the edges of each region separately
        for domain_use_tmp in domain_use:
            
            points1 = adata1_tmp.obsm[spatial_key]   # coordinates of data1
            index_tmp1 = np.where(adata1_tmp.obs[domain_key]==domain_use_tmp)[0]   # index of target domain in data1
            index_tmp_dict1 = dict(enumerate(index_tmp1))   # dict of index.  new index: raw index
            boundary_tmp1, edge_tmp1, _ = alpha_shape(points1[index_tmp1], alpha)   # detect edge of target domain 
            boundary_use_tmp1 = list(itemgetter(*boundary_tmp1)(index_tmp_dict1))    # map the boundary index back to raw index
            edge_use_tmp1 = {(index_tmp_dict1[i], index_tmp_dict1[j]) for i,j in edge_tmp1}   # combine slice edge and domain edge
            
            points2 = adata2_tmp.obsm[spatial_key]
            index_tmp2 = np.where(adata2_tmp.obs[domain_key]==domain_use_tmp)[0]
            index_tmp_dict2 = dict(enumerate(index_tmp2))
            boundary_tmp2, edge_tmp2, _ = alpha_shape(points2[index_tmp2], alpha)
            boundary_use_tmp2 = list(itemgetter(*boundary_tmp2)(index_tmp_dict2))
            edge_use_tmp2 = {(index_tmp_dict2[i], index_tmp_dict2[j]) for i,j in edge_tmp2}
        
        boundary_use_tmp1 = list(set(boundary_use_tmp1))
        boundary_use_tmp2 = list(set(boundary_use_tmp2))
        
        boundary_all.append((boundary_use_tmp1, boundary_use_tmp2))
        edge_all.append((edge_use_tmp1, edge_use_tmp2))

    return boundary_all, edge_all





import os
import numpy as np
import anndata as ad
import matplotlib.pyplot as plt

from STAIR.location.align_init import initial_alignment
from STAIR.location.align_fine import fine_alignment
from STAIR.location.edge_detection import alpha_shape, calcu_lisi, select_clustered_domains, detect_edge_of_domains
from STAIR.utils import MakeLogClass


class Loc_Align(object):
    
    """
    
    Location alignment of multiplt ST slices, including initial alignment and fine alignment. 
    They perform spatial alignmentusing similarity of spatial embedding and spatial coordinates, respectively.

    Parameters
    ----------
    adata
        AnnData object object of scanpy package
    batch_key
        The key containing slice information in .obs
    batch_order
        Slice order used to perform alignment. Align according to the default order of elements in batch_key if None

    Examples
    --------
    >>> adata = sc.read_h5ad(path_to_anndata)
    >>> loc_align = Loc_Align(adata, batch_key='batch')
    >>> loc_align.init_align(emb_key = 'HAN_SE')
    >>> loc_align.detect_edge_fine_align(domain_key = 'layer_cluster')
    >>> loc_align.plot_edge(spatial_key = 'transform_init')
    >>> adata_aligned = loc_align.fine_align()

    """

   
    def __init__(
        self,
        adata,
        batch_key,
        batch_order = None,
        make_log = True, 
        result_path = '.'
    ):
        super(Loc_Align, self).__init__()

        self.batch_key = batch_key
        
        if batch_order is None:
            batch_order = list(adata.obs[batch_key].value_counts().sort_index().index)
        
        self.batch_n = len(batch_order)
        self.adata_list = [adata[adata.obs[batch_key]==key].copy() for key in batch_order]

        self.make_log = make_log
        self.result_path = result_path
        if self.make_log:
            self.makeLog = MakeLogClass(f"{self.result_path}/log_loc.tsv").make
            
    
    def init_align(self, 
                   emb_key,
                   spatial_key = 'spatial',
                   num_mnn = 1, 
                   init_align_key = 'transform_init',
                   return_result = False
                   ):
        '''
        Initial alignment of spatial location.
        
        Parameters
        ----------
        emb_key
            AnnData object object of scanpy package
        spatial_key
            Key of raw spatial coordinates in .obsm 
        num_mnn
            The number of mutual nearest neighbors calculated according to emb_key
        init_align_key
            Key of initial transformed coordinates added in .obsm
        
        '''
        self.init_align_key = init_align_key
        self.init_adatas, anchors, self.Ts_init = initial_alignment(self.adata_list, 
                                                                    spatial_key = spatial_key,
                                                                    emb_key = emb_key,
                                                                    num_mnn = num_mnn,
                                                                    key_added = init_align_key
                                                                )
        # self.boundary = [(anchor[:,0].tolist(), anchor[:,1].tolist()) for anchor in anchors]
        
        if self.make_log:
            self.makeLog(f"Parameter set for initial alignment")
            self.makeLog(f"  Starting coordinates: {spatial_key}")
            self.makeLog(f"  K of MNN: {num_mnn}")
            self.makeLog(f"  Aligned coordinates: {init_align_key}")
        
        if return_result:
            return anchors, self.Ts_init


    def detect_fine_points( self,
                            slice_boundary = True, 
                            domain_boundary = True, 
                            domain_key = 'layer_cluster',
                            num_domains = 1,
                            sep_sort = True,
                            alpha = 70,
                            return_result = False):
    
        '''
        Prepare for fine alignment. 
        First, the spatial domain with the highest degree of spatial aggregation is selected according to the index LISI, 
        and then the edge of the slice and the aforementioned spatial domain is detected.
        
        Parameters
        ----------
        domain_key
            Key of spatial domains in .obs 
        num_domains
            Number of domains used to aligning slices.
        sep_sort
            Boolean value, whether to sort spatial clustered pattern together.
        alpha
            alpha value to detect the edge of slices and dimains.
        '''
        
        self.alpha = alpha
        self.boundary = []
        self.edge = []
        
        # detect edge of slices
        if slice_boundary:
            boundary_slices, edge_slices = [], []
            for ii in range(len(self.init_adatas)):
                if slice_boundary:
                    adata_tmp = self.init_adatas[ii]
                    spatial_tmp = adata_tmp.obsm[self.init_align_key]
                    boundary_tmp, edge_tmp, _ = alpha_shape(spatial_tmp, alpha=alpha, only_outer=True)
                else:
                    boundary_tmp, edge_tmp = [], set()
                boundary_slices.append(boundary_tmp)
                edge_slices.append(edge_tmp)
                if ii !=0 :
                    self.boundary += [(boundary_slices[ii-1], boundary_slices[ii])]
                    self.edge += [(edge_slices[ii-1], edge_slices[ii])]

        # detect edge of domains
        if domain_boundary:
            
            # calculate lisi of each domains
            lisi_list = [calcu_lisi(adata_tmp, domain_key=domain_key, spatial_key=self.init_align_key) for adata_tmp in self.init_adatas]
            
            # sort the domains according to lisi
            domains_use = [select_clustered_domains(lisi_list[i], 
                                                    lisi_list[i+1], 
                                                    domain_key, 
                                                    use_domain_nums = num_domains, 
                                                    sep_sort=sep_sort) for i in range(self.batch_n-1)]
            
            # detect edge of domains
            boundary_domain, edge_domain = detect_edge_of_domains(  self.init_adatas, 
                                                                    domain_key = domain_key, 
                                                                    domains_use = domains_use, 
                                                                    spatial_key = self.init_align_key, 
                                                                    alpha = alpha)

            for ii in range(len(self.boundary)):
                boundary_tmp = self.boundary[ii]
                boundary_tmp1, boundary_tmp2 = boundary_tmp
                boundary_tmp1 += boundary_domain[ii][0]
                boundary_tmp2 += boundary_domain[ii][1]
                boundary_tmp1 = list(set(boundary_tmp1))
                boundary_tmp2 = list(set(boundary_tmp2))
                self.boundary[ii] = (boundary_tmp1, boundary_tmp2)
                edge_tmp1, edge_tmp2 = self.edge[ii]
                edge_tmp1 = edge_tmp1.union(edge_domain[ii][0])
                edge_tmp2 = edge_tmp2.union(edge_domain[ii][1])
                self.edge[ii] = (edge_tmp1, edge_tmp2)
                
        if self.make_log:
            self.makeLog(f"Parameter set for edge detection")
            self.makeLog(f"  Spatial coordinates: {domain_key}")
            self.makeLog(f"  Number of domains: {num_domains}")
            self.makeLog(f"  Sep sort: {sep_sort}")
            self.makeLog(f"  Alpha: {alpha}")
        
        if return_result:
            if domain_boundary:
                return self.boundary, self.edge, lisi_list, domains_use
            return self.boundary, self.edge        


    def fine_align( self, 
                    fine_align_key = 'transform_fine',
                    max_iterations = 20,
                    tolerance = 1e-10,
                    return_result = False
                   ):
        
        '''
        Fine alignment of spatial location.
        
        Parameters
        ----------
        fine_align_key
            Key of fine transformed coordinates added in .obsm
        max_iterations
            Maximum number of iterations for icp
        tolerance
            Maximum error allowed for early stopping
            
        Return
        ----------
        adata_aligned
            Fine aligned adata with 'init_align_key' and 'fine_align_key' added in .obsm
        '''
        
        self.fine_adatas, Ts_fine = fine_alignment( self.init_adatas, 
                                                    self.boundary, 
                                                    spatial_key=self.init_align_key, 
                                                    key_added=fine_align_key, 
                                                    init_pose = None, 
                                                    max_iterations = max_iterations, 
                                                    tolerance = tolerance)
                    
        adata_aligned = ad.concat(self.fine_adatas)
        
        if self.make_log:
            self.makeLog(f"Parameter set for fine alignment")
            self.makeLog(f"  Starting coordinates: {self.init_align_key}")
            self.makeLog(f"  Aligned coordinates: {fine_align_key}")
            self.makeLog(f"  Max iterations: {max_iterations}")
            self.makeLog(f"  Tolerance: {tolerance}")

        if return_result:
            return adata_aligned, Ts_fine         
        return adata_aligned


    def plot_edge(self,
                  spatial_key,
                  figsize = (6,6),
                  s=1
                  ):
        
        '''
        Plot the detected edges of slices and domains to select an suitable alpha value.
        
        Parameters
        ----------
        spatial_key
            Spatial coordinates used for plot in .obsm.

        '''
        
        if spatial_key in list(self.init_adatas[0].obsm.keys()):
            adatas = self.init_adatas
        else:
            adatas = self.fine_adatas
            
        ### check edges of slices 
        for ii in range(len(self.boundary)):
            
            # get adata
            adata_tmp1 = adatas[ii]
            adata_tmp2 = adatas[ii+1]
            
            # get slices
            slice_tmp1 = list(set(adata_tmp1.obs[self.batch_key]))[0]
            slice_tmp2 = list(set(adata_tmp2.obs[self.batch_key]))[0]
            
            # get boundarys and edges
            # boundary_tmp1, boundary_tmp2 = self.boundary[ii]
            edge_tmp1, edge_tmp2 = self.edge[ii]
            
            # get coordinates of boundarys
            spatial_tmp1 = adata_tmp1.obsm[spatial_key]
            spatial_tmp2 = adata_tmp2.obsm[spatial_key]
            
            if not os.path.exists(self.result_path + '/location/edge'):
                os.makedirs(self.result_path + '/location/edge')
            
            xx,yy = np.median(spatial_tmp1, 0)
            plt.figure(figsize=figsize)
            plt.scatter(spatial_tmp1[:, 0], spatial_tmp1[:, 1], s = s)
            for i, j in edge_tmp1:
                plt.plot(spatial_tmp1[[i, j], 0], spatial_tmp1[[i, j], 1], c='#E24A33')
            plt.text(xx, yy, f"alpha={self.alpha}", size=18)
            plt.savefig(f'{self.result_path}/location/edge/spatial_edge_{slice_tmp1}_{ii}.png', bbox_inches='tight')
            plt.close()
            
            xx,yy = np.median(spatial_tmp2, 0)
            plt.figure(figsize=figsize)
            plt.scatter(spatial_tmp2[:, 0], spatial_tmp2[:, 1], s = s)
            for i, j in edge_tmp2:
                plt.plot(spatial_tmp2[[i, j], 0], spatial_tmp2[[i, j], 1], c='#8EBA42')
                # plt.plot(spatial_tmp2[[i, j], 0], spatial_tmp2[[i, j], 1], c='#988ED5')
            plt.text(xx, yy, f"alpha={self.alpha}", size=18)
            plt.savefig(f'{self.result_path}/location/edge/spatial_edge_{slice_tmp2}_{ii}.png', bbox_inches='tight')
            plt.close()
    





import numpy as np
from sklearn.neighbors import NearestNeighbors


def transform(point_cloud, T):
    ''' 
    Transform point clouds given transformation matrix
    Input:
        point_cloud: the coordinates of point clouds being transform
        T: (m+1)x(m+1) homogeneous transformation matrix. m:the dimension of coordinates.
    Output:
        point_cloud_align: the coordinates of point clouds
    '''
    
    point_cloud_align = np.ones((point_cloud.shape[0], 3))
    point_cloud_align[:,0:2] = np.copy(point_cloud)
    point_cloud_align = np.dot(T, point_cloud_align.T).T
    
    return point_cloud_align[:, :2]


def nearest_neighbor(src, dst, metric='minkowski'):
    '''
    Find the nearest (Euclidean) neighbor in dst for each point in src
    Input:
        src: Nxm array of points
        dst: Nxm array of points
    Output:
        distances: Euclidean distances of the nearest neighbor
        indices: dst indices of the nearest neighbor
    '''
    neigh = NearestNeighbors(n_neighbors=1, metric=metric)
    neigh.fit(dst)
    distances, indices = neigh.kneighbors(src, return_distance=True)
    return distances.ravel(), indices.ravel()


# https://github.com/ClayFlannigan/icp
def best_fit_transform(A, B):
    '''
    Calculates the least-squares best-fit transform that maps corresponding points A to B in m spatial dimensions
    Input:
      A: Nxm numpy array of corresponding points
      B: Nxm numpy array of corresponding points
    Returns:
      T: (m+1)x(m+1) homogeneous transformation matrix that maps A on to B
      R: mxm rotation matrix
      t: mx1 translation vector
    '''
    
    assert A.shape == B.shape
    
    # get number of dimensions
    m = A.shape[1]
   
    # translate points to their centroids
    centroid_A = np.mean(A, axis=0)
    centroid_B = np.mean(B, axis=0)
    AA = A - centroid_A
    BB = B - centroid_B
    
    # rotation matrix
    H = np.dot(AA.T, BB)
    U, S, Vt = np.linalg.svd(H)
    R = np.dot(Vt.T, U.T)
    
    # special reflection case
    if np.linalg.det(R) < 0:
       Vt[m-1,:] *= -1
       R = np.dot(Vt.T, U.T)
    
    # translation
    t = centroid_B.T - np.dot(R,centroid_A.T)
    
    # homogeneous transformation
    T = np.identity(m+1)
    T[:m, :m] = R
    T[:m, m] = t
    
    return T, R, t




import os
import numpy as np
import scanpy as sc
import matplotlib.pyplot as plt 
from sklearn.neighbors import NearestNeighbors

from STAIR.location.transformation import best_fit_transform, transform
from STAIR.location.edge_detection import alpha_shape
from STAIR.location.align_fine import fine_alignment
from STAIR.utils import MakeLogClass


def sort_slices(atte, start=None, return_tree=False):
    
    import networkx as nx
    
    matrix = 1 - (atte + atte.T)/2
    keys_use = atte.index.tolist()

    G = nx.Graph()
    for i in range(len(keys_use)):
        for j in range(len(keys_use)):
            if i == j:
                G.add_edge(keys_use[i], keys_use[j], weight=0)
            if i < j :
                G.add_edge(keys_use[i], keys_use[j], weight=matrix.iloc[i,j])

    # 计算最小生成树
    minimum_spanning_tree = nx.minimum_spanning_tree(G)

    # 输出最小生成树的边
#     for edge in minimum_spanning_tree.edges(data=True):
#         print(edge)

    result = list(minimum_spanning_tree.edges(data=True))

    if start is none:
        dists = {result[0][0]:0.} 
    else:
        dists = {start:0.} 

    un_calcu = []

    for edge in minimum_spanning_tree.edges(data=True):
        vert1, vert2, weight = edge
        weight = weight['weight']
        if vert1 in dists.keys() and vert2 in dists.keys():
            pass
        elif vert1 in dists.keys():
            dists[vert2] = dists[vert1] + weight
        elif vert2 in dists.keys():
            dists[vert1] = dists[vert2] - weight
        else:
            un_calcu.append(edge)

    while len(un_calcu)!=0:
        edge = un_calcu.pop(0)
        vert1, vert2, weight = edge
        weight = weight['weight']
        if vert1 in dists.keys() and vert2 in dists.keys():
            pass
        elif vert1 in dists.keys():
            dists[vert2] = dists[vert1] + weight
        elif vert2 in dists.keys():
            dists[vert1] = dists[vert2] - weight
        else:
            un_calcu.append(edge)

    
    if return_tree:
        return dists, minimum_spanning_tree
    return dists

def loc_predict_z(adata, 
                  atte, 
                  querys, 
                  loc_key, 
                  batch_key, 
                  knowns = None, 
                  num_mnn = 3): 
    use = adata.obs[[batch_key, loc_key]].drop_duplicates()
    use.index = use[batch_key]
    use = use[loc_key]
    if knowns is None:
        knowns = list(set(adata.obs[batch_key].value_counts().index.tolist()) - set(querys))
        knowns.sort()
    loc_knowns = use.loc[knowns] 
    preds = []
    for query_tmp in querys:
        atte_in = atte.loc[query_tmp, knowns]
        atte_out = atte.loc[knowns, query_tmp]
        neigh_in = atte_in.sort_values(ascending=False)[:num_mnn]
        neigh_out = atte_out.sort_values(ascending=False)[:num_mnn]
        neigh_index = list(set(neigh_in.index).intersection(set(neigh_out.index)))       
        pred = (atte_in[neigh_index]*loc_knowns[neigh_index]).sum() / (atte_in[neigh_index]).sum()
        preds.append(pred)
    nearest_slices = [loc_knowns.index[abs(loc_knowns - preds[i]).argmin()] for i in range(len(preds))]
    return preds, nearest_slices



def init_align_with_scale(  adata_ref,
                            adata_query, 
                            emb_key = 'HAN_SE',
                            num_mnn = 1,
                            spatial_key1 = 'spatial',
                            spatial_key2 = None,
                            use_scale = True, 
                            key_added = 'init_scale',
                            return_scale = False
                        ):
    emb1_tmp = adata_ref.obsm[emb_key]
    emb2_tmp = adata_query.obsm[emb_key]
    # print(emb1_tmp)
    # print(emb2_tmp)

    # Search for mutual nearest neighbors of two slices
    neigh1 = NearestNeighbors(n_neighbors=num_mnn, metric='cosine')
    neigh1.fit(emb2_tmp)
    indices1 = neigh1.kneighbors(emb1_tmp, return_distance=False)
    neigh2 = NearestNeighbors(n_neighbors=num_mnn, metric='cosine')
    neigh2.fit(emb1_tmp)
    indices2 = neigh2.kneighbors(emb2_tmp, return_distance=False)
    set1 = {(i, indices1[i,j]) for i in range(indices1.shape[0]) for j in range(indices1.shape[1])}
    set2 = {(indices2[j,i], j) for j in range(indices2.shape[0]) for i in range(indices2.shape[1])}
    pair = set1.intersection(set2)
    pair = np.array(list(pair))
    if spatial_key2 is None:
        spatial_key2 = spatial_key1
    B = adata_ref[pair[:,0],:].obsm[spatial_key1].copy()
    A = adata_query[pair[:,1],:].obsm[spatial_key2].copy()
    scales = np.array([np.sqrt(((B[i] - B[j])**2).sum()) / np.sqrt(((A[i] - A[j])**2).sum()) for i in range(B.shape[0]) for j in range(B.shape[0]) if i!=j])
    scale_use = np.median(scales)
    A_scaled = A * scale_use
    T,_,_ = best_fit_transform(A_scaled, B)
    # transform the coordinates of adata2
    adata_query.obsm[key_added] = transform(adata_query.obsm[spatial_key2] * scale_use, T)   
    if  key_added != spatial_key1:
        adata_ref.obsm[key_added] = adata_ref.obsm[spatial_key1].copy()
    if return_scale:
        return adata_ref, adata_query, scale_use
    return adata_ref, adata_query


def plot_edge( adata_a, adata_b, edge_tmp1, edge_tmp2, query_index, 
               alpha_query, alpha_ref, spatial_key='spatial', 
               figsize=(5,5), s_query=10, s_ref=50, result_path='.'):
    
    spatial_tmp1 = adata_a.obsm[spatial_key]
    spatial_tmp2 = adata_b.obsm[spatial_key]

    if not os.path.exists(result_path + '/loc_pred/edge'):
        os.makedirs(result_path + '/loc_pred/edge')

    xx,yy = np.median(spatial_tmp1, 0)
    plt.figure(figsize=figsize)
    plt.scatter(spatial_tmp1[:, 0], spatial_tmp1[:, 1], s = s_ref)
    for i, j in edge_tmp1:
        plt.plot(spatial_tmp1[[i, j], 0], spatial_tmp1[[i, j], 1], c='#E24A33')

    plt.text(xx, yy, f"alpha_ref={alpha_ref}", size=18)
    plt.savefig(f'{result_path}/loc_pred/edge/spatial_edge_{str(query_index)}_a.png', bbox_inches='tight')
    plt.close()

    xx,yy = np.median(spatial_tmp2, 0)
    plt.figure(figsize=figsize)
    plt.scatter(spatial_tmp2[:, 0], spatial_tmp2[:, 1], s = s_query)
    for i, j in edge_tmp2:
        plt.plot(spatial_tmp2[[i, j], 0], spatial_tmp2[[i, j], 1], c='#8EBA42')

    plt.text(xx, yy, f"alpha_query={alpha_query}", size=18)
    plt.savefig(f'{result_path}/loc_pred/edge/spatial_edge_{str(query_index)}_b.png', bbox_inches='tight')
    plt.close()



class Loc_Pred(object):
    
    """
    
    Location prediction of new slices based on known reference, including location prediction and alignment. 
    They perform location prediction and alignment based on attention and spatial embedding in HAT module.

    Parameters
    ----------
    adata
        AnnData object of scanpy package including query and reference data
    atte
        A pandas dataframe with attention as value and slice names as index and columns
    batch_key
        The key containing slice information in .obs
    querys 
        list of names of slices to predict

    Examples
    --------
    >>> adata = sc.read_h5ad(path_to_anndata)
    >>> loc_pred = Loc_Pred(adata, atte, batch_key = 'section_index', querys = ['10X_1'])
    >>> loc_pred.pred_z(loc_key = 'stereo_AP', num_mnn = 20)
    >>> ladata_query = loc_pred.pred_xy(spatial_key_query = 'spatial',
                                        spatial_key_ref = 'spatial_ccf_2d',
                                        spatial_key_3d = 'spatial_ccf_3d',
                                        emb_key = 'HAN_SE',
                                        )

    """
   
    def __init__(
        self,
        adata,
        atte, 
        batch_key,
        querys, 
        make_log = True, 
        result_path = '.'
    ):
        super(Loc_Pred, self).__init__()

        self.adata = adata
        self.atte = atte
        self.batch_key = batch_key
        self.querys = querys
        self.make_log = make_log
        self.result_path = result_path

        if self.make_log:
            self.makeLog = MakeLogClass(f"{self.result_path}/log_loc_pred.tsv").make
            self.makeLog(f"Location prediction")
            self.makeLog(f"  Slice key: {self.batch_key}")
            self.makeLog(f"  Query slices: {self.querys}")


    def pred_z( self, 
                loc_key = 'stereo_AP', 
                knowns = None, 
                num_mnn = 20, 
                return_result = True
               ):
        """
        Predict the coordinate position parallel to the slice in the reference, and find the nearest slice with the new one.
        
        Parameters
        ----------
        loc_key
            The key containing coordinates of paralle direction in .obs
        knowns
            List of slice names with known coordinates in paralle direction
        num_mnn
            Number of neighbor slices in predicting new z
        return_result
            Whether return the predicted locations and nearset slices for each query slice.
        """
        self.preds, self.nearest_slices = loc_predict_z( self.adata, 
                                                         self.atte, 
                                                         querys = self.querys, 
                                                         loc_key = loc_key, 
                                                         batch_key = self.batch_key, 
                                                         knowns = knowns, 
                                                         num_mnn = num_mnn)
        
        if self.make_log:
            self.makeLog(f"Location prediction of z")
            self.makeLog(f"  Location key: {loc_key}")
            self.makeLog(f"  Number of neighbor slices: {num_mnn}")

        if return_result:
            return self.preds, self.nearest_slices

    def pred_xy( self, 
                 spatial_key_query = 'spatial',
                 spatial_key_ref = 'spatial_ccf_2d',
                 spatial_key_init = 'spatial_init',
                 spatial_key_3d = 'spatial_ccf_3d',
                 emb_key = 'CASTLE',
                 num_mnn_init = 1,
                 return_scale = True,
                 alpha_query = 1,
                 alpha_ref = 1,
                 add_3d = True,
                 plot_init = True,
                 figsize = (5,5),
                 s_query = 10, 
                 s_ref = 50
                 ):
        """
        Get the coordinates of the new slice consistent with the reference, including initial alignment and fine alignment.

        Parameters
        ----------
        spatial_key_query
            The key of 2D coordinates of query datasets in .obsm.
        spatial_key_ref
            The key of 2D coordinates of reference in .obsm.
        spatial_key_init
            Added key of initial aligned 2D coordinates in .obsm.
        spatial_key_3d
            Added key of 3D coordinates consistent with reference in .obsm.
        emb_key
            The key containing the spatial embedding, default is 'CASTLE'.
        """

        adata_querys = []
        for i in range(len(self.querys)):
            query_slice = self.querys[i]
            nearest_slice = self.nearest_slices[i]
            adata_a = self.adata[self.adata.obs['section_index'] == nearest_slice].copy()
            adata_b = self.adata[self.adata.obs['section_index'] == query_slice].copy()

            adata_a, adata_b, scale = init_align_with_scale(adata_ref = adata_a, 
                                                            adata_query = adata_b, 
                                                            emb_key = emb_key,
                                                            num_mnn = num_mnn_init,
                                                            spatial_key1 = spatial_key_ref,
                                                            spatial_key2 = spatial_key_query,
                                                            key_added = spatial_key_init,
                                                            return_scale = return_scale)

            boundary_tmp1, edge_tmp1, _ = alpha_shape(adata_a.obsm[spatial_key_ref], alpha=alpha_ref, only_outer=True)
            boundary_tmp2, edge_tmp2, _ = alpha_shape(adata_b.obsm[spatial_key_init], alpha=alpha_query, only_outer=True)

            if plot_init:
                
                plot_edge( adata_a, adata_b, edge_tmp1, edge_tmp2, i, alpha_query, alpha_ref, 
                           spatial_key = spatial_key_init, result_path = self.result_path, 
                           figsize = figsize, s_query = s_query, s_ref = s_ref)


            fine_adatas, Ts_fine = fine_alignment(  [adata_a, adata_b], 
                                                    [(boundary_tmp1, boundary_tmp2)], 
                                                    spatial_key = spatial_key_init, 
                                                    key_added = spatial_key_ref, 
                                                    init_pose = None, 
                                                    max_iterations = 40, 
                                                    tolerance = 1e-8)

            adata_b = fine_adatas[1]
            
            if add_3d:
                adata_b.obsm[spatial_key_3d] = np.hstack((adata_b.obsm[spatial_key_ref], np.array([self.preds[i]]*adata_b.shape[0])[:,None]))
            
            adata_querys.append(adata_b)
            adata_query = sc.concat(adata_querys)
        
        if self.make_log:
            self.makeLog(f"Location prediction of x and y")
            self.makeLog(f"  2D coordinates of query: {spatial_key_query}")
            self.makeLog(f"  2D coordinates of reference: {spatial_key_ref}")
            self.makeLog(f"  2D coordinates of initial alignment: {spatial_key_init}")
            self.makeLog(f"  3D coordinates of final alignment: {spatial_key_3d}")
            self.makeLog(f"  Number of neighbor spots in initial alignment: {num_mnn_init}")
            self.makeLog(f"  Alpha of edge detection in query data: {alpha_query}")
            self.makeLog(f"  Alpha of edge detection in reference: {alpha_ref}")

        return adata_query












import numpy as np
import torch 
import matplotlib.pyplot as plt
import torch.nn.functional as F

def nb_loss(data, mean, disp):
    eps = 1e-10
    loss1 = torch.lgamma(disp+eps) + torch.lgamma(data+1) - torch.lgamma(data+disp+eps)
    loss2 = (disp+data) * torch.log(1.0 + (mean/(disp+eps))) + (data * (torch.log(disp+eps) - torch.log(mean+eps)))
    return loss1 + loss2


def zinb_loss(data, mean, disp, drop, ridge_lambda=0):
    eps = 1e-10
    nb_case = nb_loss(data, mean, disp) - torch.log(1.0-drop+eps)
    zero_nb = torch.pow(disp/(disp+mean+eps), disp)
    zero_case =  -torch.log(drop + ((1.0-drop)*zero_nb)+eps)
    result = torch.where(torch.lt(data, 1e-10), zero_case, nb_case)
    ridge = ridge_lambda*torch.pow(drop, 2)
    result += ridge
    return result.mean()


def nll_loss(data, mean, disp, drop=None, dist='zinb'):
    if dist == 'nb':
        return nb_loss(data, mean, disp).mean()
    else:
        return zinb_loss(data, mean, disp, drop, ridge_lambda=0)





import numpy as np
from .transformation import best_fit_transform, nearest_neighbor, transform

def align_fine_pair(adata1, 
                    adata2, 
                    index_use1, 
                    index_use2, 
                    spatial_key1='spatial', 
                    spatial_key2=None, 
                    key_added = 'transformed',
                    init_pose = None, 
                    max_iterations = 20, 
                    tolerance = 0.001):
    '''
    Align spatial coordinates of adata2 to match adata1 according to the specified spatial domains.
    Input:
      adata1
      adata2
      spatial_key1: key of spatial coordinates in .obsm 
      spatial_key2: key of spatial coordinates in .obsm
      domain_key: key of spatial domains in .obs 
      domain_use: domains included in 'domain_key' used to align the slice
    Returns:
      adata1 with 'transformed' in .obsm
      adata2 with 'transformed' in .obsm
      T: transformation matrix with 3 x 3
    '''
    if spatial_key2 is None:
        spatial_key2 = spatial_key1
    
    B = adata1[index_use1,:].obsm[spatial_key1]
    A = adata2[index_use2,:].obsm[spatial_key2]

    # get number of dimensions
    m = A.shape[1]

    # make points homogeneous, copy them to maintain the originals
    src = np.ones((m+1,A.shape[0]))   # (3, 238)
    dst = np.ones((m+1,B.shape[0]))   # (3, 248)
    src[:m,:] = np.copy(A.T)
    dst[:m,:] = np.copy(B.T)

    # apply the initial pose estimation
    if init_pose is not None:
        src = np.dot(init_pose, src)

    prev_error = 0
    for i in range(max_iterations): 
        
        # find the nearest neighbors between the current source and destination points
        distances, indices = nearest_neighbor(src[:m,:].T, dst[:m,:].T) 
        
        # compute the transformation between the current source and nearest destination points
        T,_,_ = best_fit_transform(src[:m,:].T, dst[:m,indices].T)
        
        # update the current source
        src = np.dot(T, src)
        
        # check error
        mean_error = np.mean(distances)
        # print('iter', i, 'error:', np.abs(prev_error - mean_error))
        
        if np.abs(prev_error - mean_error) < tolerance:
            break
        
        prev_error = mean_error
        
    # calculate final transformation
    T,_,_ = best_fit_transform(A, src[:m,:].T)

    if key_added != spatial_key1:
        adata1.obsm[key_added] = adata1.obsm[spatial_key1]
    
    adata2.obsm[key_added] = transform(adata2.obsm[spatial_key2], T)

    return adata1, adata2, T



def fine_alignment( adatas, 
                    boundarys, 
                    spatial_key='spatial', 
                    key_added='transformed', 
                    init_pose = None, 
                    max_iterations = 20, 
                    tolerance = 1e-8):
    '''
    Perform fine alignment to match the first adata in adatas according to the edge of slices and most clustered spatial domains.
    Input:
      adatas:      list of adatas to be aligned. The first adata is the fixed slice
      boundarys:   list of bounary indexs of each adata
      spatial_key: Key of spatial coordinates in .obsm
      key_added:   Key of transformed coordinates added in .obsm
      init_pose:   The initial pose estimation
    Returns:
      aligned:     adata list with key_added in .obsm added. default is 'transformed'
    '''
    
    aligned = []
    Ts = []
    
    print('Performing fine alignment of the 1 pair of data...')
    
    # fine alignment of the first two adatas
    adata1 = adatas[0]
    adata2 = adatas[1]
    
    boundary_tmp1, boundary_tmp2 = boundarys[0]

    adata1, adata2, T_tmp = align_fine_pair(adata1 = adatas[0], 
                                            adata2 = adatas[1], 
                                            index_use1=boundary_tmp1, 
                                            index_use2=boundary_tmp2, 
                                            spatial_key1=spatial_key, 
                                            spatial_key2=spatial_key, 
                                            key_added=key_added,
                                            init_pose = init_pose, 
                                            max_iterations = max_iterations, 
                                            tolerance = tolerance)
    
    aligned.append(adata1)
    aligned.append(adata2)
    Ts.append(T_tmp)
    
    # fine alignment of the rest adatas
    if len(aligned)!=len(adatas):
        for i in range(1, len(adatas)-1):
            
            print(f'Performing fine alignment of the {i+1} pair of data...')
            
            adata1_tmp = aligned[-1]
            adata2_tmp = adatas[i+1]           
            boundary_tmp1, boundary_tmp2 = boundarys[i]  
            
            for T_ in Ts:
                adata2_tmp.obsm[spatial_key] = transform(adata2_tmp.obsm[spatial_key], T_)
            
            _, adata2_tmp, T_tmp = align_fine_pair( adata1_tmp, 
                                                    adata2_tmp, 
                                                    index_use1=boundary_tmp1, 
                                                    index_use2=boundary_tmp2, 
                                                    spatial_key1=key_added, 
                                                    spatial_key2=spatial_key,
                                                    key_added = key_added, 
                                                    init_pose = init_pose, 
                                                    max_iterations = max_iterations, 
                                                    tolerance = tolerance)
            
            aligned.append(adata2_tmp)
            Ts.append(T_tmp)
    
    return aligned, Ts



def fine_alignment_(adatas, 
                    boundarys, 
                    spatial_key='spatial', 
                    key_added='transformed', 
                    init_pose = None, 
                    max_iterations = 20, 
                    tolerance = 0.001):
    '''
    Align multiple slices to match the first adata in adatas according to the specified spatial domains.
    Input:
      adatas:      list of adatas to be aligned. The first adata is the fixed slice
      boundarys:   list of bounary indexs of each adata
      spatial_key: Key of spatial coordinates in .obsm
      key_added:   Key of transformed coordinates added in .obsm
      init_pose:   The initial pose estimation
    Returns:
      aligned:     adata list with key_added in .obsm added. default is 'transformed'
    '''
    
    
    Ts_fine = []
    for i in range(len(adatas)-1):
        adata1 = adatas[i]
        adata2 = adatas[i+1]
        boundary_tmp1 = boundarys[i][0]
        boundary_tmp2 = boundarys[i][1]
        _, _, T_ = align_fine_pair( adata1, 
                                    adata2, 
                                    index_use1=boundary_tmp1, 
                                    index_use2=boundary_tmp2, 
                                    spatial_key1=spatial_key, 
                                    spatial_key2=spatial_key,
                                    key_added = key_added, 
                                    init_pose = init_pose, 
                                    max_iterations = max_iterations, 
                                    tolerance = tolerance)
        Ts_fine.append(T_)
    
    # Compute initial transformed coordinates
    for i in range(len(adatas)):          
        adatas[i].obsm[key_added] = adatas[i].obsm[spatial_key]
        if i != 0:
            T_tmp = Ts_fine[:i]
            # print(len(T_tmp))
            for T_ in T_tmp:
                adatas[i].obsm[key_added] = transform(adatas[i].obsm[key_added], T_)
    return adatas


  


import torch
from torch.utils.data import Dataset

import numpy as np
import pandas as pd
import scipy.sparse as sp
import scanpy as sc


class MyDataset(Dataset):
    '''
    construct dataset of model1
    input: adata
    output: Dataset with feature, count and library size.
    '''
    def __init__(self, adata, count_key=None, size='explog', normalize=False, hvg=False, scale=False, batch_key=None):
        super(MyDataset,self).__init__()
        
        if count_key is None:
            count_key = 'counts'
            adata.layers['counts'] = adata.X.copy()
            
        # set count matrix
        count = adata.layers[count_key].copy()
        
        if sp.issparse(count):
            self.count = torch.from_numpy(count.toarray()).float()
        else:
            self.count = torch.from_numpy(count).float() 
        
        ### library size vector
        if size == 'explog':
            self.size = torch.exp(torch.log10(self.count.sum(axis=1))).unsqueeze(1)  
        elif size == 'sum':
            self.size = self.count.sum(axis=1).unsqueeze(1)  
        elif size == 'median':
            self.size = (self.count.sum(1) / np.median(self.count.sum(1))).unsqueeze(1) 
        adata.obs['size'] = self.size.squeeze(1).numpy()
        
        # normalize 
        if normalize:
            sc.pp.normalize_per_cell(adata)
        
        # log1p
        sc.pp.log1p(adata)
        if hvg:
            sc.pp.highly_variable_genes(
                                        adata,
                                        n_top_genes=hvg,
                                        subset=True,
                                        layer=count_key,
                                        flavor="seurat_v3",
                                    )
            adata = adata[:,:hvg]
        
        # scale
        if scale:
            sc.pp.scale(adata)
        
        ### count matrix
        if sp.issparse(adata.layers[count_key]):
            self.count = torch.from_numpy(adata.layers[count_key].toarray()).float()
        else:
            self.count = torch.from_numpy(adata.layers[count_key]).float() 
        
        ### input feature matrix
        if sp.issparse(adata.X):
            self.feature = torch.from_numpy(adata.X.toarray()).float()
        else:
            self.feature = torch.from_numpy(adata.X).float()    
        
        ### batch labels
        if batch_key is not None:
            # self.batch = torch.from_numpy(adata.obs[batch_key].values)
            self.batch = torch.from_numpy(pd.get_dummies(adata.obs[batch_key]).values).float()    
            self.all_data = [(self.feature[i], self.count[i], self.size[i], self.batch[i]) for i in range(self.feature.shape[0])]
        
        else:
            self.all_data = [(self.feature[i], self.count[i], self.size[i]) for i in range(self.feature.shape[0])]
    
    def __getitem__(self,idx):
        return self.all_data[idx]      
    
    def __len__(self):
        return len(self.all_data)